%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(fun_SUITE).

-define(default_timeout, ?t:minutes(1)).

-export([all/1,init_per_testcase/2,fin_per_testcase/2,
	 bad_apply/1,bad_fun_call/1,badarity/1,equality/1,ordering/1,
	 fun_to_port/1,t_hash/1,t_phash/1,t_phash2/1,md5/1,
	 refc/1,refc_ets/1,refc_dist/1,
	 const_propagation/1,t_arity/1]).

-include("test_server.hrl").

all(suite) -> [bad_apply,bad_fun_call,badarity,equality,ordering,
	       fun_to_port,t_hash,t_phash,t_phash2,md5,
	       refc,refc_ets,refc_dist,const_propagation,
	       t_arity].

init_per_testcase(_Case, Config) ->
    ?line Dog = test_server:timetrap(?default_timeout),
    [{watchdog, Dog}|Config].

fin_per_testcase(_Case, Config) ->
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.

bad_apply(doc) ->
    "Test that the correct EXIT code is returned for all types of bad funs.";
bad_apply(suite) -> [];
bad_apply(Config) when list(Config) ->
    ?line bad_apply_fc(42, [0]),
    ?line bad_apply_fc(xx, [1]),
    ?line bad_apply_fc({}, [2]),
    ?line bad_apply_fc({1}, [3]),
    ?line bad_apply_fc({1,2,3}, [4]),
    ?line bad_apply_fc({1,2,3}, [5]),
    ?line bad_apply_fc({1,2,3,4}, [6]),
    ?line bad_apply_fc({1,2,3,4,5,6}, [7]),
    ?line bad_apply_fc({1,2,3,4,5}, [8]),
    ?line bad_apply_badarg({1,2}, [9]),
    ok.

bad_apply_fc(Fun, Args) ->
    Res = (catch apply(Fun, Args)),
    erlang:garbage_collect(),
    erlang:yield(),
    case Res of
	{'EXIT',{{badfun,Fun},_Where}} ->
	    ok = io:format("apply(~p, ~p) -> ~p\n", [Fun,Args,Res]);
	Other ->
	    ok = io:format("apply(~p, ~p) -> ~p\n", [Fun,Args,Res]),
	    ?t:fail({bad_result,Other})
    end.

bad_apply_badarg(Fun, Args) ->
    Res = (catch apply(Fun, Args)),
    erlang:garbage_collect(),
    erlang:yield(),
    case Res of
	{'EXIT',{{badfun,Fun},_Where}} ->
	    ok = io:format("apply(~p, ~p) -> ~p\n", [Fun,Args,Res]);
	Other ->
	    ok = io:format("apply(~p, ~p) -> ~p\n", [Fun,Args,Res]),
	    ?t:fail({bad_result, Other})
    end.

bad_fun_call(doc) ->
    "Try directly calling bad funs.";
bad_fun_call(suite) -> [];
bad_fun_call(Config) when list(Config) ->
    ?line bad_call_fc(42),
    ?line bad_call_fc(xx),
    ?line bad_call_fc({}),
    ?line bad_call_fc({1}),
    ?line bad_call_fc({1,2,3}),
    ?line bad_call_fc({1,2,3}),
    ?line bad_call_fc({1,2,3,4}),
    ?line bad_call_fc({1,2,3,4,5,6}),
    ?line bad_call_fc({1,2,3,4,5}),
    ?line bad_call_fc({1,2}),
    ok.

bad_call_fc(Fun) ->
    Args = [some,stupid,args],
    Res = (catch Fun(Fun(Args))),
    case Res of
	{'EXIT',{{badfun,Fun},_Where}} ->
	    ok = io:format("~p(~p) -> ~p\n", [Fun,Args,Res]);
	Other ->
	    ok = io:format("~p(~p) -> ~p\n", [Fun,Args,Res]),
	    ?t:fail({bad_result,Other})
    end.

badarity(doc) ->
    "Call and apply valid funs with wrong number of arguments.";
badarity(suite) -> [];
badarity(Config) when list(Config) ->
    ?line Fun = fun() -> ok end,
    ?line Stupid = {stupid,arguments},
    ?line Args = [some,{stupid,arguments},here],

    %% Simple call.

    ?line Res = (catch Fun(some, Stupid, here)),
    erlang:garbage_collect(),
    erlang:yield(),
    case Res of
	{'EXIT',{badarity,[{Fun,Args}|_]}} ->
	    ?line ok = io:format("~p(~p) -> ~p\n", [Fun,Args,Res]);
	_ ->
	    ?line ok = io:format("~p(~p) -> ~p\n", [Fun,Args,Res]),
	    ?line ?t:fail({bad_result,Res})
    end,

    %% Apply.

    ?line Res2 = (catch apply(Fun, Args)),
    erlang:garbage_collect(),
    erlang:yield(),
    case Res2 of
	{'EXIT',{badarity,[{Fun,Args}|_]}} ->
	    ?line ok = io:format("apply(~p, ~p) -> ~p\n", [Fun,Args,Res2]);
	_ ->
	    ?line ok = io:format("apply(~p, ~p) -> ~p\n", [Fun,Args,Res2]),
	    ?line ?t:fail({bad_result,Res2})
    end,
    ok.

equality(doc) -> "Test equality of funs.";
equality(suite) -> [];
equality(Config) when list(Config) ->
    F0 = fun() -> 1 end,
    F0_copy = copy_term(F0),
    ?line true = eq(F0, F0),
    ?line true = eq(F0, F0_copy),

    %% Compare different arities.
    F1 = fun(X) -> X + 1 end,
    ?line true = eq(F1, F1),
    ?line false = eq(F0, F1),
    ?line false = eq(F0_copy, F1),

    %% Compare different environments.
    G1 = make_fun(1),
    G2 = make_fun(2),
    ?line true = eq(G1, G1),
    ?line true = eq(G2, G2),
    ?line false = eq(G1, G2),
    ?line false = eq(G2, G1),
    G1_copy = copy_term(G1),
    ?line true = eq(G1, G1_copy),

    %% Compare fun with binaries.
    B = list_to_binary([7,8,9]),
    ?line false = eq(B, G1),
    ?line false = eq(G1, B),

    ok.

eq(X, X) -> true;
eq(_, _) -> false.

copy_term(Term) ->
    binary_to_term(term_to_binary(Term)).

make_fun(X) ->
    fun() -> X end.
	    
ordering(doc) -> "Tests ordering of funs.";
ordering(suite) -> [];
ordering(Config) when list(Config) ->
    F1 = make_fun(1, 2),
    F1_copy = copy_term(F1),
    F2 = make_fun(1, 3),
    F3 = make_fun(3, 4),

    ?line true = F1 == F1,
    ?line true = F1 == F1_copy,
    ?line true = F1 /= F2,

    ?line true = F1 < F2,
    ?line true = F2 > F1,
    ?line true = F2 < F3,
    ?line true = F3 > F2,

    ?line false = F1 > F2,
    ?line false = F2 > F3,

    %% Compare with binaries.

    B = list_to_binary([7,8,9,10]),
    ?line false = B == F1,
    ?line false = F1 == B,

    ?line true = F1 < B,
    ?line true = B > F2,

    ?line false = F1 > B,
    ?line false = B < F2,

    ?line false = F1 >= B,
    ?line false = B =< F2,

    %% Create a port and ref.

    ?line Path = ?config(priv_dir, Config),
    ?line AFile = filename:join(Path, "vanilla_file"),
    ?line P = open_port(AFile, [out]),
    ?line R = make_ref(),

    %% Compare funs with ports and refs.

    ?line true = R < F3,
    ?line true = F3 > R,
    ?line true = F3 < P,
    ?line true = P > F3,

    ?line true = R =< F3,
    ?line true = F3 >= R,
    ?line true = F3 =< P,
    ?line true = P >= F3,

    ?line false = R > F3,
    ?line false = F3 < R,
    ?line false = F3 > P,
    ?line false = P < F3,

    %% Compare funs with conses and nils.

    ?line true = F1 < [a],
    ?line true = F1 < [],
    ?line true = [a,b] > F1,
    ?line true = [] > F1,

    ?line false = [1] < F1,
    ?line false = [] < F1,
    ?line false = F1 > [2],
    ?line false = F1 > [],

    ?line false = [1] =< F1,
    ?line false = [] =< F1,
    ?line false = F1 >= [2],
    ?line false = F1 >= [],
    ok.

make_fun(X, Y) ->
    fun(A) -> A*X+Y end.

fun_to_port(doc) -> "Try sending funs to ports (should fail).";
fun_to_port(suite) -> [];
fun_to_port(Config) when list(Config) ->
    ?line fun_to_port(Config, xxx),
    ?line fun_to_port(Config, fun() -> 42 end),
    ?line fun_to_port(Config, [fun() -> 43 end]),
    ?line fun_to_port(Config, [1,fun() -> 44 end]),
    ?line fun_to_port(Config, [0,1|fun() -> 45 end]),
    B64K = build_io_list(65536),
    ?line fun_to_port(Config, [B64K,fun() -> 45 end]),
    ?line fun_to_port(Config, [B64K|fun() -> 45 end]),
    ok.

fun_to_port(Config, IoList) ->
    Path = ?config(priv_dir, Config),
    AFile = filename:join(Path, "vanilla_file"),
    Port = open_port(AFile, [out]),
    case catch port_command(Port, IoList) of
	{'EXIT',{badarg,_}} -> ok;
	Other -> ?t:fail({unexpected_retval,Other})
    end.

build_io_list(0) -> [];
build_io_list(1) -> [7];
build_io_list(N) ->
    L = build_io_list(N div 2),
    case N rem 2 of
	0 -> [L|L];
	1 -> [7,L|L]
    end.

t_hash(doc) -> "Test the hash/2 BIF on funs.";
t_hash(suite) -> [];
t_hash(Config) when list(Config) ->
    F1 = fun(_X) -> 1 end,
    F2 = fun(_X) -> 2 end,
    ?line true = hash(F1) /= hash(F2),

    G1 = make_fun(1, 2, 3),
    G2 = make_fun(1, 2, 3),
    G3 = make_fun(1, 2, 4),
    ?line true = hash(G1) == hash(G2),
    ?line true = hash(G2) /= hash(G3),
    
    ok.

hash(Term) ->
    erlang:hash(Term, 16#7ffffff).


t_phash(doc) -> "Test the phash/2 BIF on funs.";
t_phash(suite) -> [];
t_phash(Config) when list(Config) ->
    F1 = fun(_X) -> 1 end,
    F2 = fun(_X) -> 2 end,
    ?line true = phash(F1) /= phash(F2),

    G1 = make_fun(1, 2, 3),
    G2 = make_fun(1, 2, 3),
    G3 = make_fun(1, 2, 4),
    ?line true = phash(G1) == phash(G2),
    ?line true = phash(G2) /= phash(G3),
    
    ok.

phash(Term) ->
    erlang:phash(Term, 16#7ffffff).

t_phash2(doc) -> "Test the phash2/2 BIF on funs.";
t_phash2(suite) -> [];
t_phash2(Config) when list(Config) ->
    F1 = fun(_X) -> 1 end,
    F2 = fun(_X) -> 2 end,
    ?line true = phash2(F1) /= phash2(F2),

    G1 = make_fun(1, 2, 3),
    G2 = make_fun(1, 2, 3),
    G3 = make_fun(1, 2, 4),
    ?line true = phash2(G1) == phash2(G2),
    ?line true = phash2(G2) /= phash2(G3),
    
    ok.

phash2(Term) ->
    erlang:phash2(Term, 16#7ffffff).

make_fun(X, Y, Z) ->
    fun() -> {X,Y,Z} end.

md5(doc) -> "Test that MD5 bifs reject funs properly.";
md5(suite) -> [];
md5(Config) when list(Config) ->
    ContextSize = size(erlang:md5_init()),

    %% Try funs in the i/o list.
    ?line bad_md5(fun(_X) -> 42 end),
    ?line bad_md5([fun(_X) -> 43 end]),
    ?line bad_md5([1,fun(_X) -> 44 end]),
    ?line bad_md5([1|fun(_X) -> 45 end]),
    ?line B64K = build_io_list(65536),
    ?line bad_md5([B64K,fun(_X) -> 46 end]),
    ?line bad_md5([B64K|fun(_X) -> 46 end]),
    ok.
    
bad_md5(Bad) ->
    {'EXIT',{badarg,_}} = (catch erlang:md5(Bad)).

refc(Config) when list(Config) ->
    ?line F1 = fun_factory(2),
    ?line {refc,2} = erlang:fun_info(F1, refc),
    ?line F2 = fun_factory(42),
    ?line {refc,3} = erlang:fun_info(F1, refc),

    ?line process_flag(trap_exit, true),
    ?line Pid = spawn_link(fun() -> {refc,4} = erlang:fun_info(F1, refc) end),
    receive
	{'EXIT',Pid,normal} -> ok;
	Other -> ?line ?t:fail({unexpected,Other})
    end,
    ?line process_flag(trap_exit, false),
    ?line {refc,3} = erlang:fun_info(F1, refc),

    %% Garbage collect. Only the F2 fun will be left.
    ?line 7 = F1(5),
    ?line true = erlang:garbage_collect(),
    ?line 40 = F2(-2),
    ?line {refc,2} = erlang:fun_info(F2, refc),
    ok.

fun_factory(Const) ->
    fun(X) -> X + Const end.

refc_ets(Config) when list(Config) ->
    ?line F = fun(X) -> X + 33 end,
    ?line {refc,2} = erlang:fun_info(F, refc),

    refc_ets_set(F, [set]),
    refc_ets_set(F, [ordered_set]),
    refc_ets_bag(F, [bag]),
    refc_ets_bag(F, [duplicate_bag]),
    ok.

refc_ets_set(F1, Options) ->
    ?line io:format("~p", [Options]),
    ?line Tab = ets:new(kalle, Options),
    ?line true = ets:insert(Tab, {a_key,F1}),
    ?line 3 = fun_refc(F1),
    ?line [{a_key,F3}] = ets:lookup(Tab, a_key),
    ?line 4 = fun_refc(F1),
    ?line true = ets:insert(Tab, {a_key,not_a_fun}),
    ?line 3 = fun_refc(F1),
    ?line true = ets:insert(Tab, {another_key,F1}),
    ?line 4 = fun_refc(F1),
    ?line true = ets:delete(Tab),
    ?line 3 = fun_refc(F1),
    ?line 10 = F3(-23),
    ?line true = erlang:garbage_collect(),
    ?line 2 = fun_refc(F1),
    ok.

refc_ets_bag(F1, Options) ->
    ?line io:format("~p", [Options]),
    ?line Tab = ets:new(kalle, Options),
    ?line true = ets:insert(Tab, {a_key,F1}),
    ?line 3 = fun_refc(F1),
    ?line [{a_key,F3}] = ets:lookup(Tab, a_key),
    ?line 4 = fun_refc(F1),
    ?line true = ets:insert(Tab, {a_key,not_a_fun}),
    ?line 4 = fun_refc(F1),
    ?line true = ets:insert(Tab, {another_key,F1}),
    ?line 5 = fun_refc(F1),
    ?line true = ets:delete(Tab),
    ?line 3 = fun_refc(F1),
    ?line 10 = F3(-23),
    ?line true = erlang:garbage_collect(),
    ?line 2 = fun_refc(F1),
    ok.

refc_dist(Config) when list(Config) ->
    ?line {ok,Node} = start_node(fun_SUITE_refc_dist),
    ?line process_flag(trap_exit, true),
    ?line Pid = spawn_link(Node,
			   fun() -> receive
					Fun when function(Fun) ->
					    2 = fun_refc(Fun),
					    exit({normal,Fun}) end
			   end),
    ?line F = fun() -> 42 end,
    ?line 2 = fun_refc(F),
    ?line Pid ! F,
    F2 = receive
	     {'EXIT',Pid,{normal,Fun}} -> Fun;
	     Other -> ?line ?t:fail({unexpected,Other})
	 end,
    ?line 3 = fun_refc(F2),
    ?line 3 = fun_refc(F),
    refc_dist_send(Node, F).

refc_dist_send(Node, F) ->
    ?line true = erlang:garbage_collect(),
    ?line Pid = spawn_link(Node,
			   fun() -> receive
					{To,Fun} when function(Fun) ->
					    2 = fun_refc(Fun),
					    To ! Fun
				    end
			   end),
    ?line 2 = fun_refc(F),
    erlang:yield(),
    Pid ! {self(),F},
    F2 = receive
	     Fun when function(Fun) -> Fun;
	     Other -> ?line ?t:fail({unexpected,Other})
	 end,
    receive {'EXIT',Pid,normal} -> ok end,
    ?line 3 = fun_refc(F),
    ?line 3 = fun_refc(F2),
    refc_dist_reg_send(Node, F).

refc_dist_reg_send(Node, F) ->
    ?line true = erlang:garbage_collect(),
    ?line 2 = fun_refc(F),
    ?line Pid = spawn_link(Node,
			   fun() ->
				   true = register(my_fun_tester, self()),
				   receive
				       {To,Fun} when function(Fun) ->
					   2 = fun_refc(Fun),
					   To ! Fun
				   end
			   end),
    erlang:yield(),
    ?line 2 = fun_refc(F),
    {my_fun_tester,Node} ! {self(),F},
    F2 = receive
	     Fun when function(Fun) -> Fun;
	     Other -> ?line ?t:fail({unexpected,Other})
	 end,
    receive {'EXIT',Pid,normal} -> ok end,

    ?line 3 = fun_refc(F),
    ?line 3 = fun_refc(F2),
    ok.
    
fun_refc(F) ->
    {refc,Count} = erlang:fun_info(F, refc),
    Count.

const_propagation(Config) when list(Config) ->
    ?line Fun1 = fun start_node/1,
    ?line 2 = fun_refc(Fun1),
    ?line Fun2 = Fun1,
    ?line my_cmp({Fun1,Fun2}),

    ?line Fun3 = fun() -> ok end,
    ?line 2 = fun_refc(Fun3),
    ?line Fun4 = Fun3,
    ?line my_cmp({Fun3,Fun4}),
    ok.

my_cmp({Fun,Fun}) -> ok;
my_cmp({Fun1,Fun2}) ->
    io:format("Fun1: ~p", [erlang:fun_info(Fun1)]),
    io:format("Fun2: ~p", [erlang:fun_info(Fun2)]),
    ?t:fail().

t_arity(Config) when list(Config) ->
    ?line 0 = fun_arity(fun() -> ok end),
    ?line 0 = fun_arity(fun() -> Config end),
    ?line 1 = fun_arity(fun(X) -> X+1 end),
    ?line 1 = fun_arity(fun(X) -> Config =:= X end),
    A = id(42),

    %% Test that the arity is transferred properly.
    ?line process_flag(trap_exit, true),
    ?line {ok,Node} = start_node(fun_test_arity),
    ?line hello_world = spawn_call(Node, fun() -> hello_world end),
    ?line 0 = spawn_call(Node, fun(X) -> X end),
    ?line 42 = spawn_call(Node, fun(_X) -> A end),
    ?line 43 = spawn_call(Node, fun(X, Y) -> A+X+Y end),
    ?line 1 = spawn_call(Node, fun(X, Y) -> X+Y end),
    ?line 45 = spawn_call(Node, fun(X, Y, Z) -> A+X+Y+Z end),
    ok.

id(X) ->
    X.

spawn_call(Node, AFun) ->
    Pid = spawn_link(Node,
		     fun() ->
			     receive
				 {Fun,Fun,Fun} when function(Fun) ->
				     Arity = fun_arity(Fun),
				     Args = case Arity of
						0 -> [];
						_ -> lists:seq(0, Arity-1)
					    end,
				     Res = apply(Fun, Args),
				     {pid,Creator} = erlang:fun_info(Fun, pid),
				     Creator ! {result,Res}
			     end
		     end),
    Pid ! {AFun,AFun,AFun},
    Res = receive
	      {result,R} -> R;
	      Other -> ?t:fail({bad_message,Other})
	  after 10000 ->
		  ?t:fail(timeout_waiting_for_result)
	  end,
    receive
	{'EXIT',Pid,normal} -> ok;
	Other2 -> ?t:fail({bad_message_waiting_for_exit,Other2})
    after 10000 ->
	    ?t:fail(timeout_waiting_for_exit)
    end,
    Res.

fun_arity(F) ->
    {arity,Arity} = erlang:fun_info(F, arity),
    Arity.

start_node(Name) ->
    Pa = filename:dirname(code:which(?MODULE)),
    Cookie = atom_to_list(erlang:get_cookie()),
    test_server:start_node(Name, slave, 
			   [{args, "-setcookie " ++ Cookie ++" -pa " ++ Pa}]).

% stop_node(Node) ->
%    test_server:stop_node(Node).
