%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(monitor_SUITE).

-include("test_server.hrl").

-export([all/1, case_1/1, case_2/1, mon_e_1/1, demon_e_1/1, demon_1/1,
	 demon_2/1, demon_3/1, mon_1/1, mon_2/1,
	 large_exit/1, list_cleanup/1]).

-export([init_per_testcase/2, fin_per_testcase/2]).

-export([y2/1, g/1, g0/0, large_exit_sub/1]).

-export([init_per_testcase/2, fin_per_testcase/2]).

all(suite) -> [case_1, case_2, mon_e_1, demon_e_1, demon_1, mon_1,
	       mon_2, demon_2, demon_3,
	       large_exit, list_cleanup].

init_per_testcase(Func, Config) when atom(Func), list(Config) ->
    Dog=?t:timetrap(?t:minutes(1)),
    [{watchdog, Dog}|Config].

fin_per_testcase(Func, Config) ->
    Dog=?config(watchdog, Config),
    ?t:timetrap_cancel(Dog).

case_1(doc) ->
    "A monitors B, B kills A and then exits (yielded core dump)";
case_1(suite) -> [];
case_1(Config) when list(Config) ->
    ?line process_flag(trap_exit, true),
    ?line spawn_link(?MODULE, g0, []),
    ?line receive _ -> ok end,
    ok.

g0() ->
    ?line B = spawn(?MODULE, g, [self()]),
    ?line erlang:monitor(process, B),
    ?line B ! ok,
    ?line receive ok -> ok end,
    ok.

g(Parent) ->
    ?line receive ok -> ok end,
    ?line exit(Parent, foo),
    ?line ok.


case_2(doc) ->
    "A monitors B, B demonitors A (yielded core dump)";
case_2(suite) -> [];
case_2(Config) when list(Config) ->
    ?line B = spawn(?MODULE, y2, [self()]),
    ?line R = erlang:monitor(process, B),
    ?line B ! R,
    ?line receive
	      {'EXIT', _} -> ok;
	      Other ->
		  test_server:fail({rec, Other})
	  end,
    ?line expect_down(R, B, normal),
    ok.

y2(Parent) ->
    ?line R = receive T -> T end,
    ?line Parent ! (catch erlang:demonitor(R)),
    ok.

expect_down(Ref, P, Reason) ->
    receive
	{'DOWN', Ref, process, P, Reason} -> 
	    ok;
	Other ->
	    test_server:fail({rec, Other})
    end.

expect_no_msg() ->
    receive
	Msg ->
	    test_server:fail({msg, Msg})
    after 0 ->
	    ok
    end.

%%% Error cases for monitor/2

mon_e_1(doc) ->
    "Error cases for monitor/2";
mon_e_1(suite) -> [];
mon_e_1(Config) when list(Config) ->
    ?line {ok, N} = test_server:start_node(hej, slave, []),
    ?line mon_error(plutt, self()),
    ?line mon_error(process, [bingo]),
    ?line mon_error(process, {rex, N, junk}),
    ?line mon_error(process, 1),

    ?line true = test_server:stop_node(N),
    ok.

%%% We would also like to have a test case that tries to monitor something
%%% on an R5 node, but this isn't possible to do systematically.
%%%
%%% Likewise against an R6 node, which is not capable of monitoring
%%% by name, which gives a badarg on the R7 node at the call to
%%% erlang:monitor(process, {Name, Node}). This has been tested 
%%% manually at least once.

mon_error(Type, Item) ->
    case catch erlang:monitor(Type, Item) of
	{'EXIT', _} ->
	    ok;
	Other ->
	    test_server:fail({err, Other})
    end.

%%% Error cases for demonitor/1

demon_e_1(doc) ->
    "Error cases for demonitor/1";
demon_e_1(suite) -> [];
demon_e_1(Config) when list(Config) ->
    ?line {ok, N} = test_server:start_node(hej, slave, []),
    ?line demon_error(plutt, badarg),
    ?line demon_error(1, badarg),

    %% Demonitor with ref created at other node
    ?line R1 = rpc:call(N, erlang, make_ref, []),
    ?line demon_error(R1, badarg),

    %% Demonitor with ref created at wrong monitor link end
    ?line P0 = self(),
    ?line P2 = spawn(
		 fun() ->
			 P0 ! {self(), ref, erlang:monitor(process,P0)},
			 receive {P0, stop} -> ok end
		 end ),
    ?line receive 
	      {P2, ref, R2} -> 
		  ?line demon_error(R2, badarg),
		  ?line P2 ! {self(), stop};
	      Other2 ->
		  test_server:fail({rec, Other2})
	  end,

    ?line true = test_server:stop_node(N),
    ok.

demon_error(Ref, Reason) ->
    case catch erlang:demonitor(Ref) of
	{'EXIT', {Reason, _}} ->
	    ok;
	Other ->
	    test_server:fail({err, Other})
    end.

%%% No-op cases for demonitor/1

demon_1(doc) ->
    "demonitor/1";
demon_1(suite) -> [];
demon_1(Config) when list(Config) ->
    ?line true = erlang:demonitor(make_ref()),
    ok.


%%% Cases for demonitor/1

demon_2(doc) ->
    "Cases for demonitor/1";
demon_2(suite) -> [];
demon_2(Config) when list(Config) ->
    ?line R1 = erlang:monitor(process, self()),
    ?line true = erlang:demonitor(R1),
    %% Extra demonitor
    ?line true = erlang:demonitor(R1),
    ?line expect_no_msg(),

    %% Normal 'DOWN'
    ?line P2 = spawn(timer, sleep, [1]),
    ?line R2 = erlang:monitor(process, P2),
    ?line expect_down(R2, P2, normal),

    %% 'DOWN' before demonitor
    ?line P3 = spawn(timer, sleep, [100000]),
    ?line R3 = erlang:monitor(process, P3),
    ?line exit(P3, frop),
    ?line erlang:demonitor(R3),
    ?line expect_down(R3, P3, frop),

    %% Demonitor before 'DOWN'
    ?line P4 = spawn(timer, sleep, [100000]),
    ?line R4 = erlang:monitor(process, P4),
    ?line erlang:demonitor(R4),
    ?line exit(P4, frop),
    ?line expect_no_msg(),

    ok.

demon_3(doc) ->
    "Distributed case for demonitor/1 (OTP-3499)";
demon_3(suite) -> [];
demon_3(Config) when list(Config) ->
    ?line {ok, N} = test_server:start_node(hej, slave, []),

    %% 'DOWN' before demonitor
    ?line P2 = spawn(N, timer, sleep, [100000]),
    ?line R2 = erlang:monitor(process, P2),
    ?line true = test_server:stop_node(N),
    ?line true = erlang:demonitor(R2),
    ?line expect_down(R2, P2, noconnection),

    ?line {ok, N} = test_server:start_node(hej, slave, []),

    %% Demonitor before 'DOWN'
    ?line P3 = spawn(N, timer, sleep, [100000]),
    ?line R3 = erlang:monitor(process, P3),
    ?line true = erlang:demonitor(R3),
    ?line true = test_server:stop_node(N),
    ?line expect_no_msg(),

    ok.

%%% Cases for monitor/2

mon_1(doc) ->
    "Cases for monitor/2";
mon_1(suite) -> [];
mon_1(Config) when list(Config) ->
    ?line P0 = self(),
    %% Normal case
    ?line P2 = spawn(timer, sleep, [1]),
    ?line R2 = erlang:monitor(process, P2),
    ?line expect_down(R2, P2, normal),

    %% 'DOWN' with other reason
    ?line P3 = spawn(timer, sleep, [100000]),
    ?line R3 = erlang:monitor(process, P3),
    ?line exit(P3, frop),
    ?line expect_down(R3, P3, frop),

    %% Monitor fails because process is dead
    ?line R4 = erlang:monitor(process, P3),
    ?line expect_down(R4, P3, noproc),

    %% Normal case (named process)
    ?line P5 = start_jeeves(jeeves),
    ?line R5 = erlang:monitor(process, jeeves),
    ?line tell_jeeves(P5, stop),
    ?line expect_down(R5, {jeeves, node()}, normal),

    %% 'DOWN' with other reason and node explicit activation
    ?line P6 = start_jeeves(jeeves),
    ?line R6 = erlang:monitor(process, {jeeves, node()}),
    ?line tell_jeeves(P6, {exit, frop}),
    ?line expect_down(R6, {jeeves, node()}, frop),

    %% Monitor (named process) fails because process is dead
    ?line R7 = erlang:monitor(process, {jeeves, node()}),
    ?line expect_down(R7, {jeeves, node()}, noproc),

    ok.

mon_2(doc) ->
    "Distributed cases for monitor/2";
mon_2(suite) -> [];
mon_2(Config) when list(Config) ->
    ?line P0 = self(),
    ?line {ok, N1} = test_server:start_node(hej1, slave, []),

    %% Normal case
    ?line P2 = spawn(N1, timer, sleep, [4000]),
    ?line R2 = erlang:monitor(process, P2),
    ?line expect_down(R2, P2, normal),

    %% 'DOWN' with other reason
    ?line P3 = spawn(N1, timer, sleep, [100000]),
    ?line R3 = erlang:monitor(process, P3),
    ?line exit(P3, frop),
    ?line expect_down(R3, P3, frop),

    %% Monitor fails because process is dead
    ?line R4 = erlang:monitor(process, P3),
    ?line expect_down(R4, P3, noproc),

    %% Other node goes down
    ?line P5 = spawn(N1, timer, sleep, [100000]),
    ?line R5 = erlang:monitor(process, P5),

    ?line true = test_server:stop_node(N1),

    ?line expect_down(R5, P5, noconnection),

    %% Monitor fails because other node is dead
    ?line P6 = spawn(N1, timer, sleep, [100000]),
    ?line R6 = erlang:monitor(process, P6),
    ?line expect_down(R6, P6, noconnection),

    %% Start a new node that can load code in this module
    ?line PA = filename:dirname(code:which(?MODULE)),
    ?line {ok, N2} = test_server:start_node
		       (hej2, slave, [{args, "-pa " ++ PA}]),

    %% Normal case (named process)
    ?line P7 = start_jeeves({jeeves, N2}),
    ?line R7 = erlang:monitor(process, {jeeves, N2}),
    ?line tell_jeeves(P7, stop),
    ?line expect_down(R7, {jeeves, N2}, normal),

    %% 'DOWN' with other reason (named process)
    ?line P8 = start_jeeves({jeeves, N2}),
    ?line R8 = erlang:monitor(process, {jeeves, N2}),
    ?line tell_jeeves(P8, {exit, frop}),
    ?line expect_down(R8, {jeeves, N2}, frop),

    %% Monitor (named process) fails because process is dead
    ?line R9 = erlang:monitor(process, {jeeves, N2}),
    ?line expect_down(R9, {jeeves, N2}, noproc),

    %% Other node goes down (named process)
    ?line P10 = start_jeeves({jeeves, N2}),
    ?line R10 = erlang:monitor(process, {jeeves, N2}),

    ?line true = test_server:stop_node(N2),

    ?line expect_down(R10, {jeeves, N2}, noconnection),

    %% Monitor (named process) fails because other node is dead
    ?line R11 = erlang:monitor(process, {jeeves, N2}),
    ?line expect_down(R11, {jeeves, N2}, noconnection),

    ok.

%%% Large exit reason. Crashed first attempt to release R5B.

large_exit(doc) ->
    "Large exit reason";
large_exit(suite) -> [];
large_exit(Config) when list(Config) ->
    ?line f(100),
    ok.

f(0) ->
    ok;
f(N) ->
    f(),
    f(N-1).

f() ->
    ?line S0 = {big, tuple, with, [list, 4563784278]},
    ?line S = {S0, term_to_binary(S0)},
    ?line P = spawn(?MODULE, large_exit_sub, [S]),
    ?line R = erlang:monitor(process, P),
    ?line P ! hej,
    receive
	{'DOWN', R, process, P, X} ->
	    ?line io:format(" -> ~p~n", [X]),
	    if
		X == S ->
		    ok;
		true ->
		    test_server:fail({X, S})
	    end;
	Other ->
	    ?line io:format(" -> ~p~n", [Other]),
	    exit({answer, Other})
    end.

large_exit_sub(S) ->
    receive X -> ok end,
    exit(S).

%%% Testing of monitor link list cleanup
%%% by using erlang:process_info(self(), monitors)
%%% and      erlang:process_info(self(), monitored_by)

list_cleanup(doc) ->
    "Testing of monitor link list cleanup by using " ++
    "erlang:process_info/2";
list_cleanup(suite) -> [];
list_cleanup(Config) when list(Config) ->
    ?line P0 = self(),
    ?line M  = node(),
    ?line PA = filename:dirname(code:which(?MODULE)),
    ?line true = register(master_bertie, self()),

    %% Normal local case, monitor and demonitor
    ?line P1 = start_jeeves(jeeves),
    ?line {[], []} = monitors(),
    ?line expect_jeeves(P1, monitors, {monitors, {[], []}}),
    ?line R1a = erlang:monitor(process, P1),
    ?line {[{process, P1}], []} = monitors(),
    ?line expect_jeeves(P1, monitors, {monitors, {[], [P0]}}),
    ?line true = erlang:demonitor(R1a),
    ?line expect_no_msg(),
    ?line {[], []} = monitors(),
    ?line expect_jeeves(P1, monitors, {monitors, {[], []}}),
    %% Remonitor named and try again, now exiting the monitored process
    ?line R1b = erlang:monitor(process, jeeves),
    ?line {[{process, {jeeves, M}}], []} = monitors(),
    ?line expect_jeeves(P1, monitors, {monitors, {[], [P0]}}),
    ?line tell_jeeves(P1, stop),
    ?line expect_down(R1b, {jeeves, node()}, normal),
    ?line {[], []} = monitors(),

    %% Slightly weird local case - the monitoring process crashes
    ?line P2 = start_jeeves(jeeves),
    ?line {[], []} = monitors(),
    ?line expect_jeeves(P2, monitors, {monitors, {[], []}}),
    ?line {monitor_process, R2} = 
	ask_jeeves(P2, {monitor_process, master_bertie}),
    ?line {[], [P2]} = monitors(),
    ?line expect_jeeves(P2, monitors,
			{monitors, {[{process, {master_bertie, node()}}], []}}),
    ?line tell_jeeves(P2, {exit, frop}),
    timer:sleep(2000),
    ?line {[], []} = monitors(),

    %% Start a new node that can load code in this module
    ?line {ok, J} = test_server:start_node
		       (jeeves, slave, [{args, "-pa " ++ PA}]),

    %% Normal remote case, monitor and demonitor
    ?line P3 = start_jeeves({jeeves, J}),
    ?line {[], []} = monitors(),
    ?line expect_jeeves(P3, monitors, {monitors, {[], []}}),
    ?line R3a = erlang:monitor(process, P3),
    ?line {[{process, P3}], []} = monitors(),
    ?line expect_jeeves(P3, monitors, {monitors, {[], [P0]}}),
    ?line true = erlang:demonitor(R3a),
    ?line expect_no_msg(),
    ?line {[], []} = monitors(),
    ?line expect_jeeves(P3, monitors, {monitors, {[], []}}),
    %% Remonitor named and try again, now exiting the monitored process
    ?line R3b = erlang:monitor(process, {jeeves, J}),
    ?line {[{process, {jeeves, J}}], []} = monitors(),
    ?line expect_jeeves(P3, monitors, {monitors, {[], [P0]}}),
    ?line tell_jeeves(P3, stop),
    ?line expect_down(R3b, {jeeves, J}, normal),
    ?line {[], []} = monitors(),

    %% Slightly weird remote case - the monitoring process crashes
    ?line P4 = start_jeeves({jeeves, J}),
    ?line {[], []} = monitors(),
    ?line expect_jeeves(P4, monitors, {monitors, {[], []}}),
    ?line {monitor_process, R4} = 
	ask_jeeves(P4, {monitor_process, {master_bertie, M}}),
    ?line {[], [P4]} = monitors(),
    ?line expect_jeeves(P4, monitors, 
			{monitors, {[{process, {master_bertie, M}}], []}} ),
    ?line tell_jeeves(P4, {exit, frop}),
    timer:sleep(2000),
    ?line {[], []} = monitors(),
    
    %% Now, the monitoring remote node crashes
    ?line P5 = start_jeeves({jeeves, J}),
    ?line {[], []} = monitors(),
    ?line expect_jeeves(P5, monitors, {monitors, {[], []}}),
    ?line {monitor_process, R5} = 
	ask_jeeves(P5, {monitor_process, P0}),
    ?line {[], [P5]} = monitors(),
    ?line expect_jeeves(P5, monitors, 
			{monitors, {[{process, P0}], []}} ),
    ?line test_server:stop_node(J),
    timer:sleep(4000),
    ?line {[], []} = monitors(),
    
    ?line true = unregister(master_bertie),
    ok.

    


%%% Our butler for named process monitor tests

jeeves(Parent, Name, Ref) 
  when pid(Parent), atom(Name), reference(Ref) ->
    %%io:format("monitor_SUITE:jeeves(~p, ~p)~n", [Parent, Name]),
    register(Name, self()),
    Parent ! {self(), Ref},
    jeeves_loop(Parent, Name).

jeeves_loop(Parent, Name) ->
    receive
	{Parent, monitors} ->
	    Parent ! {self(), {monitors, monitors()}},
	    jeeves_loop(Parent, Name);
	{Parent, {monitor_process, P}} ->
	    Parent ! {self(), {monitor_process, 
			       catch erlang:monitor(process, P) }},
	    jeeves_loop(Parent, Name);
	{Parent, {demonitor, Ref}} ->
	    Parent ! {self(), {demonitor, catch erlang:demonitor(Ref)}},
	    jeeves_loop(Parent, Name);
	{Parent, stop} ->
	    ok;
	{Parent, {exit, Reason}} ->
	    exit(Reason);
	Other ->
	    io:format("~p:jeeves_loop received ~p~n", [?MODULE, Other])
    end.


start_jeeves({Name, Node}) 
  when atom(Name) , atom(Node) ->
    Parent = self(),
    Ref = make_ref(),
    Pid = spawn(Node, fun() -> jeeves(Parent, Name, Ref) end),
    receive 
	{Pid, Ref} -> 
	    ok;
	Other ->
	    test_server:fail({rec, Other})
    end,
    Pid;
start_jeeves(Name) ->
    start_jeeves({Name, node()}).


tell_jeeves(Pid, What) when pid(Pid) ->
    Pid ! {self(), What}.


ask_jeeves(Pid, Request) when pid(Pid) ->
    Pid ! {self(), Request},
    receive
	{Pid, Response} ->
	    Response;
	Other ->
	    test_server:fail({rec, Other})
    end.


expect_jeeves(Pid, Request, Response) when pid(Pid) ->
    Pid ! {self(), Request},
    receive 
	{Pid, Response} -> 
	    ok;
	Other ->
	    test_server:fail({rec, Other})
    end.


monitors() ->
    monitors(self()).

monitors(Pid) when pid(Pid) ->
    {monitors, Monitors}          = process_info(self(), monitors),
    {monitored_by,  MonitoredBy}  = process_info(self(), monitored_by),
    {Monitors, MonitoredBy}.

